/* Copyright © 2017 Dylan Baker
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <vector>

const int cnums[] = {0, 61};

/* Provided by foobar.c */
extern "C" int get_number_index (void);

template<typename T, int N>
std::vector<T> makeVector(const T (&data)[N])
{
    return std::vector<T>(data, data+N);
}

namespace {
    std::vector<int> numbers = makeVector(cnums);
}

extern "C" int six_one(void) {
    return numbers[get_number_index ()];
}
