// This import is needed for swiftc to implictly import the FFI module
// which will in turn conflict with the dependency's include path and error out
// if we don't manually replace all SDK paths with the newest one.
import Foundation
