// SPDX-License-Identifier: Apache-2.0
// Copyright © 2023 Intel Corporation

#include "lib.hpp"

#include <string>

namespace {

uint64_t priv_length(const std::string & str) {
    return str.length();
}

}

extern "C" uint64_t lib_length(const char * str) {
    return priv_length(str);
}
