#include "libA.h"

#undef DLL_PUBLIC
#define BUILDING_DLL
#include "config.h"

namespace meson_test_as_needed {
  namespace {
    bool set_linked() {
      linked = true;
      return true;
    }
    bool stub = set_linked();
  }

  DLL_PUBLIC int libB_unused_func() {
    return 0;
  }
}
