#!/usr/bin/env python3

import os

assert 'ENV_A' in os.environ
assert 'ENV_B' in os.environ
assert 'ENV_C' in os.environ

print('ENV_A is', os.environ['ENV_A'])
print('ENV_B is', os.environ['ENV_B'])
print('ENV_C is', os.environ['ENV_C'])
