---
short-description: The Meson reference manual
render-subpages: false
...

# Reference manual

This is the root page of the online Meson reference manual. This
manual is also available in a more machine readable format as a
JSON documented attached to every release since *0.60.0*. See our
[IDE integration](IDE-integration.md) documentation for more
information.

All functions and methods are documented in detail in the
following subpages:

## Elementary types

{{#elementary}}
{{indent}}- {{&link}}
{{/elementary}}

## Functions

The following functions are available in build files. Click on each
to see the description and usage. The objects returned by them are
[listed here](#returned-objects).

{{#functions}}
{{indent}}- {{&link}}
{{/functions}}

## Builtin objects

These are built-in objects that are always available.

{{#builtins}}
{{indent}}- {{&link}}
{{/builtins}}

## Returned objects

These are objects that can be returned by [functions](#functions)
or other methods.

{{#returned}}
{{indent}}- {{&link}}
{{/returned}}

{{#enable_modules}}
## Modules

{{#modules}}
{{indent}}- {{&link}}
{{/modules}}
{{/enable_modules}}

<!-- The links used to be generated with {>root_link}, but this is a bit hard to read  -->
